VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "NList"
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

' <todo>
' - Filter
' - Sort
' - Shuffle
' </todo>

' ******************************************************************************************
' Encapsulation
' ******************************************************************************************

Private m_First As CLink
Private m_Last As CLink
Private m_LowerBound As Long
Private m_Parent As CLink


' ******************************************************************************************
' Init/Fin
' ******************************************************************************************
Public Sub Create(ParamArray values() As Variant)
    Dim v As Variant
    For Each v In values()
        Me.Add CStr(v)
    Next v
End Sub
Public Sub SetParent(ByVal mom As CLink)
    Set m_Parent = mom
End Sub
Public Sub Dispose()
    Me.Clear
    Set m_Parent = Nothing
End Sub


' ******************************************************************************************
' Methods
' ******************************************************************************************
Public Function Add(ByVal vValue As Variant) As CLink
    Dim lnk As New CLink: lnk.Create vValue, m_Last, Nothing
    lnk.SetParent m_Parent
    If m_First Is Nothing Then Set m_First = lnk
    Set m_Last = lnk
    Set Add = lnk
End Function
Public Sub AddList(ByVal lst As NList)
    If m_First Is Nothing Then
        Set m_First = lst.First
        Set m_Last = lst.Last
    Else
        Set lst.First.PrevLink = m_Last
        Set m_Last.NextLink = lst.First
        Set m_Last = lst.Last
    End If
End Sub
Public Function Insert(ByVal vValue As Variant, ByVal index As Long) As CLink
    CheckIndex index, True ' ... speed up pos 1 insertion
    Dim lnk As New CLink
    If index = Me.LowerBound Then
        lnk.Create vValue, Nothing, m_First
        lnk.SetParent m_Parent
        Set m_First = lnk
        If m_Last Is Nothing Then Set m_Last = lnk
    ElseIf index > Me.UpperBound Then
        Set lnk = Me.Add(vValue)
    Else
        Dim prev As CLink: Set prev = Me.Item(index - 1)
        Dim naext As CLink: Set naext = Me.Item(index)
        lnk.Create vValue, prev, naext
        lnk.SetParent m_Parent
    End If
    Set Insert = lnk
End Function
Public Sub InsertList(ByVal lst As NList, ByVal index As Long)
    If m_First Is Nothing Then
        Me.AddList lst
        Exit Sub
    End If
    CheckIndex index, True ' ... speed up pos 1 insertion
    If index = Me.LowerBound Then
        Set lst.Last.NextLink = m_First
        Set m_First.PrevLink = lst.Last
        Set m_First = lst.First
    ElseIf index > Me.UpperBound Then
        Set lst.First.PrevLink = m_Last
        Set m_Last.NextLink = lst.First
        Set m_Last = lst.Last
    Else
        Dim prev As CLink: Set prev = Me.Item(index - 1)
        Dim naext As CLink: Set naext = Me.Item(index)
        Set prev.NextLink = lst.First
        Set lst.First.PrevLink = prev
        Set naext.PrevLink = lst.Last
        Set lst.Last.NextLink = naext
    End If
End Sub
Public Sub Remove(ByVal index As Long)
    CheckIndex index, False ' ... speed up pos 1 deletion
    Dim lnk As CLink: Set lnk = Me.Item(index)
    If lnk Is m_First And lnk Is m_Last Then
        Me.Clear
    ElseIf lnk Is m_First Then
        Set lnk.NextLink.PrevLink = Nothing
        Set m_First = lnk.NextLink
    ElseIf lnk Is m_Last Then
        Set lnk.PrevLink.NextLink = Nothing
        Set m_Last = lnk.PrevLink
    Else
        Set lnk.PrevLink.NextLink = lnk.NextLink
        Set lnk.NextLink.PrevLink = lnk.PrevLink
    End If
End Sub
Public Sub RemoveSpan(ByVal iFrom As Long, ByVal iTo As Long)
    CheckIndeces iFrom, iTo
    Dim lnkFrom As CLink: Set lnkFrom = Me.Item(iFrom)
    Dim lnkTo As CLink: Set lnkTo = Me.Item(iTo)
    If lnkFrom Is m_First And lnkTo Is m_Last Then
        Set m_First = Nothing
        Set m_Last = Nothing
    ElseIf lnkFrom Is m_First Then
        Set m_First = lnkTo.NextLink
        Set lnkTo.NextLink.PrevLink = Nothing
        Set lnkTo.NextLink = Nothing
    ElseIf lnkTo Is m_Last Then
        Set m_Last = lnkFrom.PrevLink
        Set lnkFrom.PrevLink.NextLink = Nothing
        Set lnkFrom.PrevLink = Nothing
    Else
        Set lnkFrom.PrevLink.NextLink = lnkTo.NextLink
        Set lnkTo.NextLink.PrevLink = lnkFrom.PrevLink
        Set lnkFrom.PrevLink = Nothing
        Set lnkTo.NextLink = Nothing
    End If
    DisposeLinks lnkFrom
End Sub
Public Sub Clear()
    Dim lnk As CLink: Set lnk = m_First
    Set m_First = Nothing
    DisposeLinks lnk
    Set m_Last = Nothing
End Sub
Public Function Cut(ByVal iFrom As Long, ByVal iTo As Long) As NList
    Set Cut = Me.Copy(iFrom, iTo)
    Me.RemoveSpan iFrom, iTo
End Function
Public Function Copy(ByVal iFrom As Long, ByVal iTo As Long) As NList
    CheckIndeces iFrom, iTo
    Dim lst As NList: Set lst = New NList
    lst.LowerBound = Me.LowerBound
    Dim lnkTo As CLink: Set lnkTo = Me.Item(iTo)
    Dim lnk As CLink: Set lnk = Me.Item(iFrom)
    lst.Add lnk.Value
    Do Until lnk Is lnkTo
        Set lnk = lnk.NextLink
        lst.Add lnk.Value
    Loop
    Set Copy = lst
End Function
Public Sub Shift(ByVal oldIndex As Long, ByVal newIndex As Long)
    CheckIndex newIndex, False ' need check before removal
    Dim lnk As CLink: Set lnk = Me.Item(oldIndex)
    Me.Remove oldIndex
    Me.Insert lnk.Value, newIndex
End Sub
Public Sub ShiftSpan(ByVal iFrom As Long, ByVal iTo As Long, ByVal newIndex As Long)
    If newIndex < Me.LowerBound Then Err.Raise 9
    If newIndex + iTo - iFrom > Me.UpperBound Then Err.Raise 9
    Dim lst As NList: Set lst = Me.Cut(iFrom, iTo)
    Me.InsertList lst, newIndex
End Sub
Public Sub Reverse()
    Dim lnk As CLink, lnkTemp As CLink
    Set lnk = m_First
    Do Until lnk Is Nothing
        Set lnkTemp = lnk.NextLink
        Set lnk.NextLink = lnk.PrevLink
        Set lnk.PrevLink = lnkTemp
        Set lnk = lnkTemp
    Loop
    Set lnkTemp = m_First
    Set m_First = m_Last
    Set m_Last = lnkTemp
End Sub


' ******************************************************************************************
' Properties
' ******************************************************************************************
Public Property Get First() As CLink
    Set First = m_First
End Property
Public Property Get Last() As CLink
    Set Last = m_Last
End Property
Public Property Get Count() As Long
    Dim lnk As CLink: Set lnk = m_First
    Do Until lnk Is Nothing
        Count = Count + 1
        Set lnk = lnk.NextLink
    Loop
End Property
Public Property Get LowerBound() As Long
    LowerBound = m_LowerBound
End Property
Public Property Let LowerBound(ByVal lb As Long)
    m_LowerBound = lb
End Property
Public Property Get UpperBound() As Long
    UpperBound = Me.Count + Me.LowerBound - 1
End Property


' ******************************************************************************************
' Element access
' ******************************************************************************************
Public Property Get Item(ByVal index As Long) As CLink
    Dim lnk As CLink: Set lnk = m_First
    Dim i As Long: i = Me.LowerBound
    Do Until lnk Is Nothing Or i > index
        If i = index Then
            Set Item = lnk
            Exit Do
        End If
        Set lnk = lnk.NextLink
        i = i + 1
    Loop
    If Item Is Nothing Then Err.Raise 9
End Property
Public Function GetEnumerator() As NListWalker
    Set GetEnumerator = New NListWalker
    GetEnumerator.Attach Me
End Function


' ******************************************************************************************
' Helpers
' ******************************************************************************************
Private Sub CheckIndex(ByVal index As Long, ByVal fInsert As Boolean)
    Dim nGrow As Long: If fInsert = True Then nGrow = 1
    Dim nUBound As Long: nUBound = Me.UpperBound
    If index < Me.LowerBound Or index > nUBound + nGrow Then Err.Raise 9
End Sub
Private Sub CheckIndeces(ByVal iFrom As Long, ByVal iTo As Long)
    If iFrom > iTo Then Err.Raise 9
    Dim nUBound As Long: nUBound = Me.UpperBound
    If iFrom < Me.LowerBound Or iFrom > nUBound Then Err.Raise 9
    If iTo < Me.LowerBound Or iTo > nUBound Then Err.Raise 9
End Sub
Private Sub DisposeLinks(ByRef lnkFrom As CLink)
    Dim lnkNext As CLink
    Do Until lnkFrom Is Nothing
        Set lnkNext = lnkFrom.NextLink
        lnkFrom.Dispose
        Set lnkFrom = lnkNext
    Loop
End Sub
